#encoding: UTF-8 
=begin
/***************************************************************************
 *   ©2019-2019 Michael Uplawski <michael.uplawski@uplawski.eu>            *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
=end

require 'nokogiri'
require 'zip'
require_relative 'logging'
require_relative 'translating'

# A sheet in a spreadsheet workbook
class Sheet
  extend Logging
  include Translating
  @@log = init_logger(STDOUT)

  Known_Expressions = ['< ', '> ', '<= ', '>= ', 'empty', 'not empty', 'equal to', 'not ' ]

  def initialize(sheet, zip_entry, table, shared_strings)
    @log = @@log
    @shared_strings = shared_strings
    # zip_entry is just a marker for the time, which allows the spreadsheet
    # to locate this sheet.
    @zip_entry = zip_entry
    @log.debug('creating sheet  ' << sheet)
    # find the file for this sheet
    node = Nokogiri::XML::fragment(sheet).at_xpath(".//sheet")
    @name = node.attribute('name').value
    # beware that xml attribute-values are Strings.
    sheetId = node.attribute('sheetId').value
    @id = sheetId.to_i
    # Its XML-content
    @table = Nokogiri::XML::Document.parse(table)
    # flag to negate expressions like
    # '... value in cell 1 is not empty' => '!cell.value.empty?'
    @negate = false;
  end

  def to_s
    @table.to_s if @table 
  end

  # Delete the rows where the value in the referenced cell produces a true
  # statement. expr must be one of the known expressions, comprising a
  # numerical value, if any. 
  def remove_line_be(reference, expr)
    @log.debug('original expression is ' << expr)
    expression = match_expression(expr)
    exit false if(!expression)
    negi = (@negate ? '!' : '' )
    expr_format = "%s (%s %s )" 

    @log.debug('full expression is ' << negi << expression)
    # Find the rows to delete
    # Translate value to the index of its entry in sharedStrings.xml, if possible.
    value = shared_string_index(value)
    # All rows
    rows = all_rows
    # @log.debug('rows is ' << rows.to_s )
    del_rows = Array.new
    rows.each do |row| 
      cell = row.css('c').detect {|c| c.attribute('r').value().start_with?(reference.upcase)  }
      if cell
        cell_value =  cell.at_css('v')
        cell_value = (cell_value ? cell_value.text : "''")
        @log.debug("cell_value is " << cell_value.to_s)
        # Use the Ruby-Force to evaluate the String expression on the cell-value
        comparison = expr_format %[negi, cell_value, expression]
        if eval comparison  
          # note the deleted row for later.
          del_rows.push(row)
          # unlink the row from its xml-context
          row.remove 
          @log.debug('He was only a ♫♪♬ ... but one more is gone!')
        end
      end
    end
    # ... later:
    # Remove the already unlinked nodes from the node-list.
    # I cannot say that I comprehend why this is necessary.
    delete_from_nodeset(all_rows, del_rows)
    adjust_rows
    @log.debug('Table is now ' << @table.to_s)
  end

  # Delete the rows which contain the given value in the referenced cell.
  # This is the same as remove_line_be(reference, '= value')
  def remove_line_bv(reference, value)
    # Find the rows to delete
    # Translate value to the index of its entry in sharedStrings.xml, if possible.
    value = shared_string_index(value)
    del_rows = Array.new
    @log.debug('have ' << all_rows.length.to_s << ' rows')
    all_rows.each do |row| 
      # take aim
      if (row.css('c').detect{|c| c.attribute('r').value().start_with?(reference.upcase) && c.at_css('v') && c.at_css('v').text == value.to_s} )
        # ... shoot
        row.remove 
        # ... move the body out of the way
        del_rows.push(row)
      end
    end
    # ... later:
    # Remove the already unlinked nodes from the node-list.
    # I cannot say that I comprehend why this is necessary.
    delete_from_nodeset(all_rows, del_rows)
    adjust_rows
    @log.debug('have now ' << all_rows.length.to_s << ' rows')
    @log.debug('Table is now ' << @table.to_s)
  end

  attr_reader :table, :name, :id, :zip_entry

  private

  def delete_from_nodeset(nodeset, nodes) 
    nodes.each { |n| nodeset.delete(n) }
  end

  # Correct row-numbers and formulas
  # i.e. eliminate empty lines.
  def adjust_rows
    all_rows.each_with_index do |row, i|
      row_ref = (i+1).to_s
      # @log.debug('row index ' << row.attribute('r') )
      row.set_attribute('r', row_ref) 
      row.children.filter('c').each do |c|
        # identify the row-attribute of the cells which needs adjustment
        ref = c.attribute('r')
        # the column value which will be altered
        col = ref.value.match(/\d+/)[0]
        # ... then do it.
        # This looks like planches and nails. Works though.
        c.set_attribute('r', ref.value.gsub!(col, row_ref) )
      end
      #@log.debug('row is now' << row.to_s)
    end
    # no forumlas yet...
  end

  # Nokogiri and xpath are fine. Xpath and me are fine, too.
  # But Nokogiri and xpath and me do not get along well...
  # I am working on that. It must though be css for the time.
  def all_rows
    @table.css("row")  
  end

  # Known_Expressions = ['< ', '> ', '<= ', '>= ', 'empty', 'not empty', 'equal to', '= ', '!= ', 'not equal to'  ]
  def match_expression(expr)
    known_expr = Known_Expressions.detect {|e| expr.start_with?(e)}
    if (!known_expr)
      @log.error( (trl('The expression %s cannot be interpreted') %expr) << '!')
      @log.error(trl('Valid expressions are') << ':')
      @log.error("\n\t" << Known_Expressions.join("\n\t"))
      @log.error(trl("Joined expressions are not (yet) supported") << '.')
      return nil
    end
    if(expr.start_with?('not') )
      # ... HA !
      @negate = true
    end   
    expr = '.to_s.empty?' if(expr.start_with?('empty') || expr.start_with?('not empty') )
    expr = "== " << shared_string_index(expr.split.last).to_s if (expr.start_with?('equal') || expr.start_with?('not equal')) 

    return expr
  end

  def shared_string_index(value)
    if value.respond_to?(:to_str)
      strings = @shared_strings.css('t') 
      string = strings.css('t').detect{|s| s.text == value }
      index = strings.index(string) 
      value = index if index
    end
    return value
  end

end

############# TEST ###############

if __FILE__ == $0
  #TODO
end
